##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
#adapted from various trunk xmlrpc files jwashin 2005-06-09

from minjson import read, write, ReadException, WriteException

__docformat__ = 'restructuredtext'

from StringIO import StringIO
import sys, traceback
from zLOG import LOG, INFO
from cgi import FieldStorage

#set Debug to True for writing messages to the console
Debug = False

response_content_type = 'text/xml'
# XXX should put this to text/x-json, but ONLY
# if the js code is patched, too!
# and even now it breaks Opera. So for now leave it on text/xml.
#response_content_type = 'text/x-json'

def parse_input(data):
    """Parse input data and return a method path and argument tuple

    The data is a string.
    """

    if Debug:
        print "processing inputs (%s)" % data
    data = read(data)
    functionstr = data[u'method'].encode('utf-8')
    if functionstr:
        method = functionstr.replace('.', '/')
    else:
        method = None
    jsonID = data[u'id']
    params = data[u'params']
    # Translate '.' to '/' in meth to represent object traversal.
    method = method.replace('.', '/')
    return jsonID, method, params

class Response:
    '''Customized Response
    '''
    # use delegation, rather than inheritance

    def __init__(self, real, jsonID): 
        self.__dict__['_real'] = real
        self.__dict__['_jsonID'] = jsonID
        real._encode_unicode = self._encode_unicode

    def __getattr__(self, name): return getattr(self._real, name)
    def __setattr__(self, name, v): return setattr(self._real, name, v)
    def __delattr__(self, name): return delattr(self._real, name)

    def setBody(self, body, title='', is_error=0, bogus_str_search=None):
        """return
        {
        'id' : matches id in request
        'result' : the result or null if error
        'error' : the error or null if result
        }
        """
        if self._jsonID is None:
            #self._real.setHeader('content-type', response_content_type)
            self._real.setBody('')
            self._real.setStatus(204)
        else:
            body = premarshal(body)
            wrapper = {'id':self._jsonID}
            wrapper['result'] = body
            wrapper['error'] = None
            if Debug:
                print wrapper
            self._real.setHeader('content-type', response_content_type)
            self._real.setBody(write(wrapper))
            self._real.setStatus(200)
        return self

    def exception(self, fatal=0, info=None,
                  absuri_match=None, tag_search=None):
        # Fetch our exception info. t is type, v is value and tb is the
        # traceback object.
        if type(info) is type(()) and len(info)==3:
            t, v, tb = info
        else: 
            t, v, tb = sys.exc_info()

        if Debug:
            for file, lineno, function, text in traceback.extract_tb(tb):
                print '%s line %s in %s' % (file, lineno, function)
                print "=>%s" % repr(text)
            print "** %s: %s" % (t, v)
        s = '%s: %s' % (getattr(t, '__name__', t), v)
        wrapper = {'id': self._jsonID}
        wrapper['result'] = None
        wrapper['error'] = s
        
        self._real.setHeader('content-type', response_content_type)
        self._real.setBody(write(wrapper))
        self._real.setStatus(200)

        #TODO What should really happen on error in a notification?
        
response=Response

def premarshal_dict(data):
    """return a non-proxied dict"""
    return dict([(premarshal(k), premarshal(v))
                 for (k, v) in data.items()])

def premarshal_list(data):
    """return a non-proxied list"""
    return map(premarshal, data)

#note: no dates or datetimes in json, but supported by xmlrpc
premarshal_dispatch_table = {
    dict: premarshal_dict,
    list: premarshal_list,
    tuple: premarshal_list,
    }

premarshal_dispatch = premarshal_dispatch_table.get

def premarshal(data):
    premarshaller = premarshal_dispatch(data.__class__)
    if premarshaller is not None:
        return premarshaller(data)
    return data

# --
# Patching processInputs of ZPublisher.HTTPRequest
# --

def processInputs(self, **kw):
    'Process request inputs'
    response=self.response
    environ=self.environ
    method=environ.get('REQUEST_METHOD','GET')

    if method != 'GET': fp=self.stdin
    else:               fp=None

    form=self.form
    other=self.other

    meth=None
    fs=FieldStorage(fp=fp,environ=environ,keep_blank_values=1)

    # Fix broken content types of Opera
    # this is because the xmlHTTPRequest method cannot set
    # the headers so it will always send text/xml;... ,
    # but we want to identify the requests regardless!
    if fs.headers.has_key('content-type') and \
            fs.headers['content-type'].startswith('text/xml') and \
            isinstance(fs.value, str) and fs.value.startswith('{"id":'):
        fs.headers['content-type'] = 'text/x-json'

    if fs.headers.has_key('content-type') and \
            fs.headers['content-type'] == 'text/x-json' \
            and method == 'POST':
        jsonID, meth, self.args = parse_input(fs.value)
        response = Response(response, jsonID)
        other['RESPONSE'] = self.response = response
        self.maybe_webdav_client = 0
        # continue with what used to be at the end of processInputs
        if environ.has_key('PATH_INFO'):
            path=environ['PATH_INFO']
            while path[-1:]=='/': path=path[:-1]
        else: path=''
        other['PATH_INFO']=path="%s/%s" % (path,meth)
        self._hacked_path=1
        return self
    else:
        self.stdin.seek(0)
        return self._processInputs_jsonrc_patched(**kw)

from ZPublisher.HTTPRequest import HTTPRequest
def patch_HTTPRequest():
    'This will patch HTTPRequest to enable json-rpc handling'
    HTTPRequest._processInputs_jsonrc_patched, HTTPRequest.processInputs = \
        HTTPRequest.processInputs, processInputs
    LOG('jsonserver', INFO, '*** Patching ZPublisher.HTTPRequest for json-rpc ***')
